/*-
 * Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2009 May 08 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: ipfw_get.c,v 1.12 2011/12/01 18:27:57 babolo Exp $";
#endif /* not lint */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/sysctl.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <net/if.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <babolo/BLINflag.h>

#define IPFW_INTERNAL
#if __FreeBSD_version < 500000
# include <netinet/ip_fw2.h>
# include <net/route.h>
# include <sys/mbuf.h>
#else
# include <netinet/ip_fw.h>
#endif
#include <netinet/ip_dummynet.h>
#ifndef DN_IS_PIPE
# include "old_dn.h"
#endif

#include "bsdfig.h"
#include "bsdfig_local.h"

static int
inisock(bsdfig_conf *cnf) {
    int ex = EX_OK;

    if  ((cnf->iprsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        ifBLIN_QV1(cnf->flags) warn("inisock: socket");
        ex = EX_OSERR;
    }
    return(ex);
}

static int
dosockopt(bsdfig_conf *cnf, int optname, void **buf, socklen_t *size, size_t delta) {
    socklen_t len, try;
    u_int32_t flocal;
    int ex = EX_OK;

    flocal = !*buf;
    len = *size;
    if  (len < delta) {
        ifBLIN_QV1(cnf->flags) warnx("dosockopt: size %u < delta %"BLIN_U, len, BLIN_I(delta));
        errno = EINVAL;
        ex = EX_SOFTWARE;
    } else {
        ex = inisock(cnf);
    }
    for (try = len; !ex; try *= 2) {
        len = try - delta;
        if  (flocal) {
            free(*buf);
            if  (!(*buf = malloc(try))) {
                ifBLIN_QV1(cnf->flags) warnx("dosockopt: no mem %d", try);
                errno = ENOMEM;
                ex = EX_OSERR;
                break;
        }   }
        if  ((ex = getsockopt(cnf->iprsock, IPPROTO_IP, optname, *buf + delta, &len))) {
            ifBLIN_QV2(cnf->flags) warn( "dosockopt: getsockopt(%d, %d, *, %"BLIN_U")"
                                       , cnf->iprsock, optname, try - delta
                                       );
            ex = EX_OSERR;
            break;
        }
        len = len + delta;
        ifBLIN_QV5(cnf->flags)
            fprintf(stderr, "dosockopt try=%u len=%u delta=%u ex=%d\n", try, len, delta, ex);
        if  (len < try) break;
    }
    *size = len;
    return(ex);
}

bsdfig_dummynets *
bsdfig_read_dummynets(bsdfig_conf *cnf) {
#   define delta (offsetof(typeof(*r.ret), dummies))
#   define FIRST_DUMMY_SIZE PAGE_SIZE
    int ex = EX_OK;
    socklen_t len;
    union {
        bsdfig_dummynets *ret;
        void *r;
    } r;

    r.ret = NULL;
    if  (!cnf) errx(EX_USAGE, "bsdfig_read_dummynets: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_dummynets %08X\n", cnf->flags);
    len = FIRST_DUMMY_SIZE;
    if  (len < delta) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_read_dummynets: len %u < %"BLIN_U, len, BLIN_I(delta));
        errno = EINVAL;
        ex = EX_SOFTWARE;
    } else if ((ex = dosockopt(cnf, IP_DUMMYNET_GET, &r.r, &len, delta))) {
        ifBLIN_QV2(cnf->flags) warn("bsdfig_read_dummynets: dosockopt");
    }
    if  (ex) {
        free(r.ret);
        r.ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_dummynets - %d\n", ex);
    } else {
        r.ret->size = len - delta;
        ifBLIN_QV5(cnf->flags) {
            unsigned i;

            for (i = 0; i < r.ret->size; i++) {
                if  (!(i & 0xF)) fprintf(stderr, "\n%08X:", i);
                if  (!(i & 0x7)) fprintf(stderr, " ");
                if  (!(i & 0x1)) fprintf(stderr, " ");
                fprintf(stderr, "%02X", ((u_char *)&(r.ret->dummies))[i]);
            }
            fprintf(stderr, "\n");
        }
        ifBLIN_QV4(cnf->flags)
            fprintf(stderr, "-bsdfig_read_dummynets [%"BLIN_U"]\n", BLIN_I(r.ret->size));
    }
    return(r.ret);
#   undef delta
}

bsdfig_ipfwrules *
bsdfig_read_ipfwrules(bsdfig_conf *cnf) {
#   define delta (offsetof(typeof(*r.ret), rules))
#   define FIRST_RULE_SIZE PAGE_SIZE
    int ex = EX_OK;
    socklen_t len;
    union {
        bsdfig_ipfwrules *ret;
        void *r;
    } r;

    r.ret = NULL;
    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwrules: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwrules %08X\n", cnf->flags);
    len = FIRST_RULE_SIZE;
    if  (len < delta) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_read_dummynets: len %u < %"BLIN_U, len, BLIN_I(delta));
        errno = EINVAL;
        ex = EX_SOFTWARE;
    } else if ((ex = dosockopt( cnf, IP_FW_GET, &r.r, &len, delta))) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwrules: dosockopt");
    }
    if  (ex) {
        free(r.ret);
        r.ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwrules - %d\n", ex);
    } else {
        r.ret->size = (len - delta) / sizeof(ipfw_insn);
        ifBLIN_QV4(cnf->flags)
            fprintf(stderr, "-bsdfig_read_ipfwrules [%"BLIN_U"]\n", BLIN_I(r.ret->size));
    }
    return(r.ret);
#   undef delta
}

#if __FreeBSD_version >= 700000
bsdfig_ipfwnats *
bsdfig_read_ipfwnats(bsdfig_conf *cnf) {
#   define delta (offsetof(typeof(*r.ret), nat_cnt))
#   define FIRST_NAT_SIZE PAGE_SIZE
    int ex = EX_OK;
    socklen_t len;
    union {
        bsdfig_ipfwnats *ret;
        void *r;
    } r;

    r.ret = NULL;
    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwnats: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwnats %08X\n", cnf->flags);
    len = FIRST_NAT_SIZE;
    if  (len < delta) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_read_dummynets: len %u < %"BLIN_U, len, BLIN_I(delta));
        errno = EINVAL;
        ex = EX_SOFTWARE;
    } else if ((ex = dosockopt(cnf, IP_FW_NAT_GET_CONFIG, &r.r, &len, delta))) {
        ifBLIN_QV2(cnf->flags) warn("bsdfig_read_ipfwnats: dosockopt");
    }
    if  (ex) {
        free(r.ret);
        r.ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnats - %d\n", ex);
    } else {
        r.ret->size = len - delta;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnats [%"BLIN_U"]\n", BLIN_I(len));
    }
    return(r.ret);
#   undef delta
}

bsdfig_ipfwnatl *
bsdfig_read_ipfwnatl(bsdfig_conf *cnf) {
#   define delta (offsetof(typeof(*r.ret), natl))
#   define FIRST_NAT_SIZE PAGE_SIZE
    int ex = EX_OK;
    socklen_t len;
    union {
        bsdfig_ipfwnatl *ret;
        void *r;
    } r;

    r.ret = NULL;
    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwnatl: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwnatl %08X\n", cnf->flags);
    len = FIRST_NAT_SIZE;
    if  (len < delta) {
        ifBLIN_QV1(cnf->flags) warnx("bsdfig_read_dummynets: len %u < %"BLIN_U, len, BLIN_I(delta));
        errno = EINVAL;
        ex = EX_SOFTWARE;
    } else if ((ex = dosockopt(cnf, IP_FW_NAT_GET_LOG, &r.r, &len, delta))) {
        ifBLIN_QV2(cnf->flags) warn("bsdfig_read_ipfwnatl: dosockopt");
    }
    if  (ex) {
        free(r.ret);
        r.ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnatl - %d\n", ex);
    } else {
        r.ret->size = len - delta;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwnatl [%"BLIN_U"]\n", BLIN_I(len));
    }
    return(r.ret);
#   undef delta
}
#endif

bsdfig_ipfwtables *
bsdfig_read_ipfwtables(bsdfig_conf *cnf) {
    bsdfig_ipfwtables *ret = NULL;
    int ex = EX_OK;
    socklen_t len, try;
    size_t i;

    if  (!cnf) errx(EX_USAGE, "bsdfig_read_ipfwtables: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_read_ipfwtables %08X\n", cnf->flags);
    if  (!(ret = malloc(cnf->tabn * sizeof(ipfw_table*) + sizeof(bsdfig_ipfwtables)))) {
        ifBLIN_QV1(cnf->flags) warnx( "bsdfig_read_ipfwtables: no mem %"BLIN_U
                                    , cnf->tabn * sizeof(ipfw_table*) + sizeof(bsdfig_ipfwtables)
                                    );
        errno = ENOMEM;
        ex = EX_OSERR;
    } else {
        bzero(&(ret->tables), ret->tabn * sizeof(ipfw_table*) + sizeof(bsdfig_ipfwtables));
        ret->tabn = cnf->tabn;
        ex = inisock(cnf);
    }
    for (i = 0; !ex && (i < ret->tabn); i++) {
        u_int32_t tbn;
        tbn = i;
        len = sizeof(tbn);
        if  ((ex = getsockopt(cnf->iprsock, IPPROTO_IP, IP_FW_TABLE_GETSIZE, &tbn, &len))) {
            ifBLIN_QV1(cnf->flags) warn("bsdfig_read_ipfwtables: getsockopt");
            ex = EX_OSERR;
            break;
        } else if (len != sizeof(tbn)) {
            ifBLIN_QV1(cnf->flags)
                warn("bsdfig_read_ipfwtables: IP_FW_TABLE_GETSIZE len %u <> %u", len, sizeof(tbn));
            errno = EBADMSG;
            ex = EX_OSERR;
            break;
        } else if (tbn) {
            tbn++;
            for (try = tbn * sizeof(ipfw_table_entry) + sizeof(ipfw_table); !ex; try *= 2) {
                if  (!(ret->tables[i] = malloc(try))) {
                    ifBLIN_QV1(cnf->flags) warnx("bsdfig_read_ipfwtables: no mem %d", try);
                    errno = ENOMEM;
                    ex = EX_OSERR;
                    break;
                }
                bzero(ret->tables[i], sizeof(ipfw_table));
                ret->tables[i]->tbl = i;
                ex = getsockopt(cnf->iprsock, IPPROTO_IP, IP_FW_TABLE_LIST, ret->tables[i], &try);
                if  (ex) {
                    ifBLIN_QV2(cnf->flags)
                        warn( "bsdfig_read_ipfwtables: getsockopt(%d, %d, %d, *, %"BLIN_U")=%d"
                            , cnf->iprsock, IPPROTO_IP, IP_FW_TABLE_LIST, try, ex
                            );
                    ex = EX_OSERR;
                    break;
                }
                ifBLIN_QV5(cnf->flags) fprintf( stderr, "[%u] %u %u %u\n"
                                              , try
                                              , ret->tables[i]->size
                                              , ret->tables[i]->cnt
                                              , ret->tables[i]->tbl
                                              );
                if  (ret->tables[i]->cnt < ret->tables[i]->size) break;
                /* IP_FW_TABLE_LIST     size    */
                /*      ,  ,        */
                /*    ,                    */
    }   }   }
    if  (ex) {
        if  (ret) for (i = 0; i < ret->tabn; i++) free(ret->tables[i]);
        free(ret);
        ret = NULL;
        ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_read_ipfwtables - %d\n", ex);
    } else {
        ifBLIN_QV4(cnf->flags)
            fprintf(stderr, "-bsdfig_read_ipfwtables [%"BLIN_U"]\n", BLIN_I(ret->tabn));
    }
    return(ret);
}

static const size_t mule[3] = {1024U, 512U, 512U};

bsdfig_conf *
bsdfig_init(u_int32_t flags, FILE *out, bsdfig_conf *slug) {
    bsdfig_conf *cnf;
    int ex = EX_OK, nv;
    size_t ns, i;

    ifBLIN_QV4(flags) fprintf(stderr, "+bsdfig_init %08X\n", flags);
    do {
        if  (slug) {
            cnf = slug;
        } else if (!(cnf = calloc(1, sizeof(bsdfig_conf)))) {
            ifBLIN_QV1(flags) warnx("bsdfig_init: no init mem %"BLIN_U, BLIN_I(sizeof(bsdfig_conf)));
            errno = ENOMEM;
            ex = EX_OSERR;
            break;
        } else {
            cnf->flags = flags;
            cnf->iprsock = -1;
        }
        if  (!cnf->out) cnf->out = out;
        if  (!cnf->syntax) cnf->syntax = &bsdfig_s;
        if  (!cnf->fibn) {
            ns = sizeof(int);
            if  (sysctlbyname("net.fibs", &nv, &ns, NULL, 0) < 0) {
                ifBLIN_QV1(cnf->flags) warn("bsdfig_init: sysctlbyname");
                ex = EX_OSERR;
                break;
            }
            cnf->fibn = nv;
        }
        if  (!cnf->tabn) cnf->tabn = IPFW_TABLES_MAX;
        if  (!(BSDFIG_0CONF & cnf->flags)) break;

        if  (!cnf->conf) {
            if  (!( cnf->conf
                  = calloc(1, (cnf->tabn - 1) * sizeof(mular_descriptor*) + sizeof(bsdfig_allconf))
                ) ) {
                ifBLIN_QV1(flags) warnx("bsdfig_init: no mem %"BLIN_U, sizeof(bsdfig_allconf));
                errno = ENOMEM;
                ex = EX_OSERR;
                break;
            } else cnf->conf->flags = cnf->flags;
        }

        if  (!cnf->conf->fibn) cnf->conf->fibn = cnf->fibn;
        if  (!cnf->conf->tabn) cnf->conf->tabn = cnf->tabn;

        if  (!cnf->conf->dummynets) {
            if  (!(cnf->conf->dummynets = mular_create(0, 3, 0, mule))) {
                ifBLIN_QV1(flags) warn("bsdfig_init: mular_create for dummynets");
                ex = EX_SOFTWARE;
                break;
            } else cnf->conf->dummynets->tofree  = free;
        }
      
        if  (!cnf->conf->ipfwrules) {
            if  (!(cnf->conf->ipfwrules = mular_create(0, 3, 0, mule))) {
                ifBLIN_QV1(flags) warn("bsdfig_init: mular_create for ipfwrules");
                ex = EX_SOFTWARE;
                break;
            } else cnf->conf->ipfwrules->tofree  = free;
        }

        if  (!cnf->conf->ipfwtables) {
            if  (!( cnf->conf->ipfwtables
                  = calloc(1, cnf->conf->tabn * sizeof(mular_descriptor*))
                ) ) {
                ifBLIN_QV1(flags)
                    warnx("bsdfig_init: no mem %"BLIN_U, cnf->conf->tabn * sizeof(mular_descriptor*));
                errno = ENOMEM;
                ex = EX_OSERR;
                break;
        }   }

        for (i = 0; !ex && (i < cnf->conf->tabn); i++) {
            if  (!cnf->conf->ipfwtables[i]) {
                if  (!(cnf->conf->ipfwtables[i] = mular_create(0, 3, 0, mule))) {
                    ifBLIN_QV1(flags) warn("bsdfig_init: mular_create for ipfwtables");
                    ex = EX_SOFTWARE;
                    break;
                } else cnf->conf->ipfwtables[i]->tofree  = free;
        }   }

        if  (0 && !cnf->conf->routes) { /* No routes yet */
            if  (!( cnf->conf->routes
                  = calloc(1, cnf->conf->fibn * sizeof(void*))
                ) ) {
                ifBLIN_QV1(flags)
                    warnx("bsdfig_init: no mem %"BLIN_U, cnf->conf->fibn * sizeof(void*));
                errno = ENOMEM;
                ex = EX_OSERR;
                break;
        }   }

        if  (ex || !(BSDFIG_MYCNF & flags)) break;
        for (i = 0; i < (cnf->tabn / BSDFIG_BASELEN); i++) cnf->conf->my_ipfwtables[i] = ~0LLU;
        for (i = 0; i < (cnf->tabn % BSDFIG_BASELEN); i++) {
            cnf->conf->my_ipfwtables[cnf->tabn / BSDFIG_BASELEN] |= 1LLU << i;
        }
        memset(&(cnf->conf->my_pipes    ), ~0, BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BYTELEN);
        memset(&(cnf->conf->my_queues   ), ~0, BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BYTELEN);
        memset(&(cnf->conf->my_ipfwrules), ~0, BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BYTELEN);
        cnf->conf->my_pipes    [0] = 0xFFFFFFFFFFFFFFFELLU;
        cnf->conf->my_queues   [0] = 0xFFFFFFFFFFFFFFFELLU;
        cnf->conf->my_ipfwrules[0] = 0xFFFFFFFFFFFFFFFELLU;
        cnf->conf->my_ipfwrules[(BSDFIG_IPFW_OTHERS_MAX / BSDFIG_BASELEN) - 1] = 0x7FFFFFFFFFFFFFFFLLU;
    } while (0);
    ifBLIN_QV4(flags) fprintf(stderr, "-bsdfig_init %d\n", ex);
    return(cnf);
}
